/*************************************************************\
|* prlmatch.c - PARALLEL IMPLEMENTATION TO MATCHING PROBLEM. *|
\*************************************************************/

#include "graph.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <sys/time.h>

extern int NUM_PROCESSORS; 

mutex_t aug_block;                    /* Control the mutual exclusion in
                                         the augmenting path routine.
                                         Only one CPU can augment a path at
                                         a time. */
word erro = 0;                        /* Count the number of times that a 
                                         search had to be done again */
word im = 0,                          /* initial matching                */
     pm = 0,                          /* parallel matching               */
     mm = 0;                          /* maximum matching                */
hrtime_t t_hrv_im_start [MAX_THREADS],/* time spent by initial matching  */
         t_hrv_im_end   [MAX_THREADS];
hrtime_t t_hrv_start [MAX_THREADS],   /* time spent by parallel matching */
         t_hrv_end   [MAX_THREADS];
hrtime_t start,                       /* start counting time             */
         end_im,                      /* end of initial matching         */
         start_seq,                   /* start sequential part           */
         end,                         /* end of the algorithm            */
         max_hrv_hp,                  /* max hrv initial matching        */
         max_hrv_pm;                  /* max hrv parallel matching       */

long num_task [MAX_THREADS];


/* Used to pass parameters to threads */
 
typedef struct {
  
  graph *graph_ptr;

  word begin,   /* boundary to work safely (begin...end) */
       
       end,
       
       index;   /* index of each thread    (0...CPU - 1) */
	
} thr_param;


/* get_free_vert - get a free vertex from the free vertex list
   input - the graph structure
   output - a free vertex if it existe or zero otherwise */

word get_free_vert (graph *graph_ptr) {

  word vert = 0;

  mutex_lock (&graph_ptr->mut_free_vert);

  if (graph_ptr->free_vert_ptr) {

    free_vert *free_vert_aux = graph_ptr->free_vert_ptr;

    graph_ptr->free_vert_ptr = graph_ptr->free_vert_ptr->next;

    vert = free_vert_aux->vert;

    free (free_vert_aux);

  }

  mutex_unlock (&graph_ptr->mut_free_vert);

  return vert;

}

/* insert_bfs - insert the vertex into the circular list
   input - the search structure and the vertex */

void insert_bfs (search *search_ptr, word vert) {

  /* the vertex is already into the list */

  if (IS_BFS_ACTIVE (search_ptr->flag_ptr [vert].flag))

    return;


  BFS_ACTIVE (search_ptr->flag_ptr [vert].flag);

  search_ptr->bfs_ptr [search_ptr->end].vert = vert;

  search_ptr->end++;

  if (search_ptr->end > search_ptr->size)

    search_ptr->end = 0;

}


/* get_bfs - get a vertex from the circular list
   input - the search structure 
   output  - the next vertex of the bread-first search
   obs - return 0 (zero) if the list is empty */

word get_bfs (search *search_ptr) {

  word vert;


  if (search_ptr->begin == search_ptr->end)

    return (0);

  vert = search_ptr->bfs_ptr [search_ptr->begin].vert;

  search_ptr->begin++;

  if (search_ptr->begin > search_ptr->size)

    search_ptr->begin = 0;

  BFS_INACTIVE (search_ptr->flag_ptr [vert].flag);

  return (vert);

}


/* reverse_path - change the directions of the predecessors pointers
                  between high and low.
   input - the searche structure the vertex high and the low */

void reverse_path (search *search_ptr, word high, word low) {

  word vert_1 = high,

       vert_2,

       vert_3;


  if (vert_1 != low) {

    vert_2 = search_ptr->pred_ptr [vert_1].pred;

    num_task [search_ptr->index]++;

    vert_3 = search_ptr->pred_ptr [vert_2].pred;

    num_task [search_ptr->index]++;

    search_ptr->pred_ptr [vert_2].pred = vert_1;

    while (vert_2 != low) {

      vert_1 = vert_2;

      vert_2 = vert_3;

      vert_3 = search_ptr->pred_ptr [vert_2].pred;

      num_task [search_ptr->index]++;

      search_ptr->pred_ptr [vert_2].pred = vert_1;

    }

  }

}


/* open_blossom - find the path through blossoms 
   input - the search strucuture, the initial vertex belonging
           to a blossom B and a pointer to that blossom 
   output - the base*(B) */

word open_blossom (search *search_ptr, word vert, blossom *blossom_ptr) {

  if (IS_OUTER (search_ptr->flag_ptr [vert].flag))

    /* The path goes directly to the base of the blossom */

    find_path (search_ptr, vert,
	       search_ptr->bl_ptr [vert].bl->base, blossom_ptr);

  else {

    /* the path passes through the peaks of the blossom */

    word left_peak = search_ptr->bl_ptr [vert].bl->left_peak,

	 right_peak = search_ptr->bl_ptr [vert].bl->right_peak;


    if (IS_RIGHT (search_ptr->flag_ptr [vert].flag)) {

      find_path (search_ptr, right_peak, vert, blossom_ptr);

      find_path (search_ptr,
		 left_peak, search_ptr->bl_ptr [vert].bl->base, blossom_ptr);

      reverse_path (search_ptr, right_peak, vert);

      search_ptr->pred_ptr [right_peak].pred = left_peak;

    }

    else {

      find_path (search_ptr, left_peak, vert, blossom_ptr);

      find_path (search_ptr, right_peak,
                 search_ptr->bl_ptr [vert].bl->base, blossom_ptr);

      reverse_path (search_ptr, left_peak, vert);

      search_ptr->pred_ptr [left_peak].pred = right_peak;

    }

  }

  return (search_ptr->bl_ptr [vert].bl->base);

}


/* find_path - find the path between the vertices high and low 
   input - the search structure, the vertices high and low and a pointer
           to a blossom indicating that a path through that blossom must
           be found */

void find_path (search *search_ptr,
                word high, word low, blossom *blossom_ptr) {

  word vert = high;


  while (vert != low)

    /* if blossom (vert) != NULL && blossom (vert) != B */

    if (search_ptr->bl_ptr [vert].bl != blossom_ptr)

      vert = open_blossom (search_ptr, vert, search_ptr->bl_ptr [vert].bl);

    else {

      vert = search_ptr->pred_ptr [vert].pred;

      num_task [search_ptr->index]++;

    }

}


/* augment_path - match the unmatched edges of the path and unmatch the others
   input - the graph, the search strucutres and the vertices high and low
   output - TRUE if the path was augmented, FALSE otherwise */

boole augment_path (graph *graph_ptr, search *search_ptr,
	 	    word high, word low) {

  word vert = high;

  word pred;


  mutex_lock (&aug_block);

  /* Verify if the path is an augmenting path */

  if (MATCHED (high) || MATCHED (low)) {

     erro++;
	 
     mutex_unlock (&aug_block);
	
     return (FALSE);
		       
  }

  vert = search_ptr->pred_ptr [high].pred;

  num_task [search_ptr->index]++;

  pred = search_ptr->pred_ptr [vert].pred;

  num_task [search_ptr->index]++;

  while (vert != low) {

    if (graph_ptr->mate_ptr [vert].mate != pred ||
        graph_ptr->mate_ptr [pred].mate != vert) {

      erro++;

      mutex_unlock (&aug_block);

      return (FALSE);

    }

    vert = search_ptr->pred_ptr [pred].pred;

    num_task [search_ptr->index]++;

    pred = search_ptr->pred_ptr [vert].pred;

    num_task [search_ptr->index]++;

  }

  /* Augment the path */

  vert = high;

  pred = search_ptr->pred_ptr [vert].pred;

  num_task [search_ptr->index]++;

  graph_ptr->mate_ptr [vert].mate = pred;

  graph_ptr->mate_ptr [pred].mate = vert;

  num_task [search_ptr->index]++;

  while (pred != low) {

    vert = search_ptr->pred_ptr [pred].pred;

    num_task [search_ptr->index]++;

    pred = search_ptr->pred_ptr [vert].pred;

    num_task [search_ptr->index]++;

    graph_ptr->mate_ptr [vert].mate = pred;

    graph_ptr->mate_ptr [pred].mate = vert;

    num_task [search_ptr->index]++;
  } 

  mutex_unlock (&aug_block);

  return (TRUE);

}


/* base - return the base*(blossom (vert))
   input - the search structure and the vertex vert
   output - base*(blossom (vert)) */

word base (search *search_ptr, word vert) {

  word v = vert;


  while (search_ptr->bl_ptr [v].bl) 

    v = search_ptr->bl_ptr [v].bl->base;

  return (v);

}


/* create_blossom - identify the vertices of the blossom and
                    create a new one to the search strucuture
   input  - the search structure and the vertices s and t */

void create_blossom (search *search_ptr, word s, word t) {

  word vert_left,
  
       vert_right,
       
       ant_left,
       
       ant_right;

  blossom *blossom_ptr;


  if (search_ptr->bl_ptr [s].bl)

    vert_left = base (search_ptr, s);

  else

    vert_left = s;


  if (search_ptr->bl_ptr [t].bl)

    vert_right = base (search_ptr, t);

  else

    vert_right = t;


  if (vert_left == vert_right)

    return;


  /* Create a blossom into the list of blossoms */

  blossom_ptr = ALLOC (1, blossom);

  blossom_ptr->left_peak = s;

  blossom_ptr->right_peak = t;

  blossom_ptr->next = search_ptr->blossom_ptr;

  search_ptr->blossom_ptr = blossom_ptr;

 
  /* Look for the vertices of the blossom */

  LEFT (search_ptr->flag_ptr [vert_left].flag);

  ant_left = vert_left;


  RIGHT (search_ptr->flag_ptr [vert_right].flag);

  ant_right = vert_right;


  while (vert_left != vert_right)

    if (search_ptr->level_ptr [vert_left].level >=
	search_ptr->level_ptr [vert_right].level) {

      search_ptr->bl_ptr [ant_left].bl = blossom_ptr;

      /* Mark the odd vertices an put them again into the circular list */

      if (IS_ODD (search_ptr->flag_ptr [ant_left].flag)) {

        EVEN (search_ptr->flag_ptr [ant_left].flag);

        insert_bfs (search_ptr, ant_left);

      }

      vert_left = search_ptr->pred_ptr [vert_left].pred;

      num_task [search_ptr->index]++;

      if (search_ptr->bl_ptr [vert_left].bl)

	vert_left = base (search_ptr, vert_left);

      ant_left = vert_left;

      LEFT (search_ptr->flag_ptr [vert_left].flag);

    } else {

      search_ptr->bl_ptr [ant_right].bl = blossom_ptr;

      /* Mark the odd vertices an put them again into the circular list */

      if (IS_ODD (search_ptr->flag_ptr [ant_right].flag)) {

        EVEN (search_ptr->flag_ptr [ant_right].flag);

        insert_bfs (search_ptr, ant_right);

      }

      vert_right = search_ptr->pred_ptr [vert_right].pred;

      num_task [search_ptr->index]++;

      if (search_ptr->bl_ptr [vert_right].bl)

	vert_right = base (search_ptr, vert_right);

      ant_right = vert_right;

      RIGHT (search_ptr->flag_ptr [vert_right].flag);

    }


  /* vertex_left is de DEEPEST COMMOM VERTEX (DCV) */

  blossom_ptr->base = vert_left; /* DCV */

}


/* search_algorithm - perform the search locally at the thread 
   input - the graph structure, the searche structure and the root of 
           the search */

boole search_algorithm (graph *graph_ptr, search *search_ptr, word root) {

  word vert, adj;


  while (vert = get_bfs (search_ptr)) {

    if (IS_EVEN (search_ptr->flag_ptr [vert].flag)) {

      edge *edge_ptr = graph_ptr->vert_ptr [vert].edge_ptr;


      while (edge_ptr) {

	adj = edge_ptr->vert;


        num_task [search_ptr->index]++;

        if (search_ptr->pred_ptr [vert].pred != adj) {

	  if (VERT_FREE (adj) && adj != root) {
            
            if (!IS_VISITED (search_ptr->flag_ptr [adj].flag)) {

              /* Try to augment the path */

  	      search_ptr->pred_ptr [adj].pred = vert;

	      find_path (search_ptr, adj, root, NULL);

	      return (augment_path (graph_ptr, search_ptr, adj, root));

	    }

          } else {

	    if (IS_VISITED (search_ptr->flag_ptr [adj].flag)) {

              if (IS_EVEN (search_ptr->flag_ptr [adj].flag)) {

                if (search_ptr->bl_ptr [vert].bl !=
                    search_ptr->bl_ptr [adj].bl ||
                    !search_ptr->bl_ptr [vert].bl) {

                  /* May put an odd label to a vertex with an even label
                     this is the condition to be a blossom */

                  create_blossom (search_ptr, vert, adj);

                }  

              }

            } else {

              /* Expand the search through the edges */

  	      VISITED (search_ptr->flag_ptr [adj].flag);

	      ODD (search_ptr->flag_ptr [adj].flag);

	      INNER (search_ptr->flag_ptr [adj].flag);

	      search_ptr->level_ptr [adj].level =

	      search_ptr->level_ptr [vert].level + 1;

	      search_ptr->pred_ptr [adj].pred = vert;

	      insert_bfs (search_ptr, adj);

            }

          }

        } /* if pred */

       	edge_ptr = edge_ptr->next;

      } /* end while adj */

    } else { /* ODD */

      adj = graph_ptr->mate_ptr [vert].mate;


      num_task [search_ptr->index]++;

      if (search_ptr->pred_ptr [vert].pred != adj) {

        if (IS_VISITED (search_ptr->flag_ptr [adj].flag)) { 

          if (IS_ODD (search_ptr->flag_ptr [adj].flag)) {

            if (search_ptr->bl_ptr [vert].bl !=
                search_ptr->bl_ptr [adj].bl ||
                !search_ptr->bl_ptr [vert].bl) {

              /* May put an odd label to a vertex with an even label
                 this is the condition to be a blossom */

              create_blossom (search_ptr, vert, adj);

            }

          }

        } else { /* Vertex not visited */ 

          /* Just expande the search to the vertex */

    	  VISITED (search_ptr->flag_ptr [adj].flag);

  	  EVEN (search_ptr->flag_ptr [adj].flag);

	  OUTER (search_ptr->flag_ptr [adj].flag);

	  search_ptr->level_ptr [adj].level =

	  search_ptr->level_ptr [vert].level + 1;

	  search_ptr->pred_ptr [adj].pred = vert;

	  insert_bfs (search_ptr, adj);

        }

      } else /* end if pred */ { /* unexpected situation, the search must
                                    be done again */
	 erro++;

	 return (FALSE);

      }

    } /* end ODD */

  } /* end While getBFS */

  return (TRUE);

}


/* parallel_maximal_matching - each thread makes a greedy matching in the
                               boundary informed
   input - the index of the processor,
	   the graph, and the boundary to work (begin...end) */

void parallel_maximal_matching (word index, graph *graph_ptr, word begin, word end) {

  word vert;


  for (vert = begin; vert <= end; vert++) 

    if (VERT_FREE (vert)) {

      edge *edge_ptr = graph_ptr->vert_ptr [vert].edge_ptr;


      while  (edge_ptr) {

        word adj = edge_ptr->vert;


        num_task [index]++;

        if (VERT_FREE (adj)) {

          graph_ptr->mate_ptr [vert].mate = adj;

          graph_ptr->mate_ptr [adj].mate = vert;

          break;

        }

        edge_ptr = edge_ptr->next;

      }

    }

}


/* parallel_matching - the function that will become a thread 
   input - a structure thr_param */

void *parallel_matching (void *input_ptr) {

  thr_param *input = (thr_param *) input_ptr;

  graph *graph_ptr = input->graph_ptr;

  word index = input->index,
  
       begin = input->begin,
      
       end = input->end;

  search *search_ptr = ALLOC (1, search);

  free_vert *free_vert_begin = NULL,

            *free_vert_end = NULL;


  word count,

       vert;


  search_ptr->index = index;

  t_hrv_im_start [index] = gethrvtime ();
 
  parallel_maximal_matching (index, graph_ptr, begin, end);

  sync ();

  if (!index)

    im = statistic_graph (graph_ptr);

  t_hrv_im_end [index] = gethrvtime ();


  /* Build the local list of free vertices */

  for (count = begin; count <= end; count++)

    if (VERT_FREE (count)) {

      free_vert *free_vert_aux = ALLOC (1, free_vert);

      free_vert_aux->vert = count;

      if (!free_vert_end)

        free_vert_end = free_vert_aux;

      free_vert_aux->next = free_vert_begin;

      free_vert_begin = free_vert_aux;

    }


  /* Put the local list into the global list */

  if (free_vert_begin) {

    mutex_lock (&graph_ptr->mut_free_vert);

    free_vert_end->next = graph_ptr->free_vert_ptr;

    graph_ptr->free_vert_ptr = free_vert_begin;

    mutex_unlock (&graph_ptr->mut_free_vert);

  }

  sync ();

  if (!index)

    end_im = gethrtime ();

  t_hrv_start [index] = gethrvtime ();


  search_ptr->size = graph_ptr->num_vertices;

  search_ptr->level_ptr = ALLOC (graph_ptr->num_vertices + 1, level);

  search_ptr->pred_ptr = ALLOC (graph_ptr->num_vertices + 1, pred);

  search_ptr->flag_ptr = ALLOC (graph_ptr->num_vertices + 1, flag);

  search_ptr->bl_ptr = ALLOC (graph_ptr->num_vertices + 1, bl);

  search_ptr->bfs_ptr = ALLOC (graph_ptr->num_vertices + 1, bfs);


  while (vert = get_free_vert (graph_ptr)) {

    if (VERT_FREE (vert)) {

      boole done = FALSE;

      word init;

      flag *flag_ptr = search_ptr->flag_ptr;

      bl *bl_ptr = search_ptr->bl_ptr;

      blossom *blossom_ptr;

      do {

          for (init = 0; init <= graph_ptr->num_vertices; init++) {

  	    flag_ptr [init].flag = 0;

	    bl_ptr [init].bl = NULL;

          }

          search_ptr->begin = search_ptr->end = 0;

          search_ptr->pred_ptr [vert].pred = vert;

          search_ptr->level_ptr [vert].level = 0;

          VISITED (flag_ptr [vert].flag);

          EVEN (flag_ptr [vert].flag);

          OUTER (flag_ptr [vert].flag);

          insert_bfs (search_ptr, vert);

          done = search_algorithm (graph_ptr, search_ptr, vert);


          /* free the memory used by the blossoms */

          blossom_ptr = search_ptr->blossom_ptr;

          while (blossom_ptr) {

            blossom *blossom_ptr_aux = blossom_ptr;

  	    blossom_ptr = blossom_ptr->next;

	    free (blossom_ptr_aux);

          }

          search_ptr->blossom_ptr = NULL;

      } while (!done && VERT_FREE (vert)); /* Prevent from expanding a 
					      search from a matched 
					      vertex */
    }

  }

  free (search_ptr->level_ptr);

  free (search_ptr->pred_ptr);

  free (search_ptr->flag_ptr);

  free (search_ptr->bl_ptr);

  free (search_ptr->bfs_ptr);

  free (search_ptr);

  t_hrv_end [index] = gethrvtime ();

  thr_exit (0);

}

/* sequential_matching - garantee of maximum matching
   input - the graph */

void sequential_matching (graph *graph_ptr) {

  search *search_ptr = ALLOC (1, search);

  word vert;


  search_ptr->size = graph_ptr->num_vertices;

  search_ptr->level_ptr = ALLOC (graph_ptr->num_vertices + 1, level);

  search_ptr->pred_ptr = ALLOC (graph_ptr->num_vertices + 1, pred);

  search_ptr->flag_ptr = ALLOC (graph_ptr->num_vertices + 1, flag);

  search_ptr->bl_ptr = ALLOC (graph_ptr->num_vertices + 1, bl);

  search_ptr->bfs_ptr = ALLOC (graph_ptr->num_vertices + 1, bfs);

  for (vert = 1; vert <= graph_ptr->num_vertices; vert++) {

    if (VERT_FREE (vert)) {

      word init;

      flag *flag_ptr = search_ptr->flag_ptr;

      bl *bl_ptr = search_ptr->bl_ptr;

      blossom *blossom_ptr;


      for (init = 0; init <= graph_ptr->num_vertices; init++) {

	flag_ptr [init].flag = 0;

	bl_ptr [init].bl = NULL;

      }


      search_ptr->begin = search_ptr->end = 0;

      search_ptr->pred_ptr [vert].pred = vert;

      search_ptr->level_ptr [vert].level = 0;

      VISITED (flag_ptr [vert].flag);

      EVEN (flag_ptr [vert].flag);

      OUTER (flag_ptr [vert].flag);

      insert_bfs (search_ptr, vert);

      search_algorithm (graph_ptr, search_ptr, vert);

      /* free the memory used by the blossoms */

      blossom_ptr = search_ptr->blossom_ptr;

      while (blossom_ptr) {

        blossom *blossom_ptr_aux = blossom_ptr;

	blossom_ptr = blossom_ptr->next;

	free (blossom_ptr_aux);

      }

      search_ptr->blossom_ptr = NULL;

    }

  }

  free (search_ptr->level_ptr);

  free (search_ptr->pred_ptr);

  free (search_ptr->flag_ptr);

  free (search_ptr->bl_ptr);

  free (search_ptr->bfs_ptr);

  free (search_ptr);

  return;

}


/* start_threads - start the threads and mark the time spent to
                   solve the matching 
   input - the graph structure */

void start_threads (graph *graph_ptr) {

  thr_param thr_input [MAX_THREADS];

  void *input_ptr [MAX_THREADS];

  word count;

  word intervall = graph_ptr->num_vertices / NUM_PROCESSORS;

  long tot_task = 0;


  for (count = 0; count < NUM_PROCESSORS; count++) 

    num_task [count] = 0;

  /* prepare the informations of the arguments */

  for (count = 0; count < NUM_PROCESSORS - 1; count++) {

    thr_input [count].graph_ptr = graph_ptr;

    thr_input [count].index = count;

    thr_input [count].begin = intervall * count + 1;

    thr_input [count].end = intervall * (count + 1);

  }

  thr_input [NUM_PROCESSORS - 1].graph_ptr = graph_ptr;

  thr_input [NUM_PROCESSORS - 1].index = NUM_PROCESSORS - 1;

  thr_input [NUM_PROCESSORS - 1].begin = intervall * (NUM_PROCESSORS - 1) + 1;

  thr_input [NUM_PROCESSORS - 1].end = graph_ptr->num_vertices;

  /* make the array of voids point to the arguments */

  for (count = 0; count < NUM_PROCESSORS; count++) 

    input_ptr [count] = &thr_input [count];

  sync_init (NUM_PROCESSORS);

  start = gethrtime ();

  mutex_init (&aug_block, USYNC_THREAD, NULL);

  if (create_threads (NUM_PROCESSORS, parallel_matching, input_ptr)) {

    fprintf (stderr, "Couldn't create threads\n");

    exit (ERROR);

  }

  join ();

  start_seq = gethrtime ();

  pm = statistic_graph (graph_ptr);

  /*sequential_matching (graph_ptr);*/

  mm = statistic_graph (graph_ptr);

  end = gethrtime ();

  end_of_sync ();

  for (count = 0; count < NUM_PROCESSORS; count++) 

    t_hrv_start [count] = t_hrv_end [count] - t_hrv_start [count];

  max_hrv_pm = t_hrv_start [0];

  max_hrv_hp = t_hrv_im_start [0];

  for (count = 1; count < NUM_PROCESSORS; count++) {

    if (max_hrv_pm < t_hrv_start [count])

      max_hrv_pm = t_hrv_start [count];

    if (max_hrv_hp < t_hrv_im_start [count])

      max_hrv_hp = t_hrv_im_start [count];

  }

  for (count = 0; count < NUM_PROCESSORS; count++) 

    tot_task += num_task [count];

  printf ("M %d CPU %d bim (%2.3f, %2.3f) im %d%% bmp (%2.3f, %2.3f) pm %d Tot %2.3f Tsks %ld Err %d\n",
    mm / 2 ,
    NUM_PROCESSORS,
    (float) max_hrv_hp /1e9,
    (float) (end_im - start)/1e9,
    im * 100 / graph_ptr->num_vertices,
    (float) max_hrv_pm /1e9,
    (float) (start_seq - end_im)/1e9,
    pm / 2,
    (float) (start_seq - start)/1e9,
    tot_task,
    erro);

  return;

}
















